/**
 * Back 2 Browser Bytecode Translator
 * Copyright (C) 2012 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://opensource.org/licenses/GPL-2.0.
 */
package org.apidesign.vm4brwsr;

import java.io.IOException;
import java.io.InputStream;
import org.apidesign.javap.AnnotationParser;
import org.apidesign.javap.ClassData;
import org.apidesign.javap.FieldData;
import org.apidesign.javap.MethodData;
import org.apidesign.javap.StackMapIterator;
import static org.apidesign.javap.RuntimeConstants.*;

/** Translator of the code inside class files to JavaScript.
 *
 * @author Jaroslav Tulach <jtulach@netbeans.org>
 */
public abstract class ByteCodeToJavaScript {
    private ClassData jc;
    private final Appendable out;

    protected ByteCodeToJavaScript(Appendable out) {
        this.out = out;
    }
    
    /* Collects additional required resources.
     * 
     * @param internalClassName classes that were referenced and should be loaded in order the
     *   generated JavaScript code works properly. The names are in internal 
     *   JVM form so String is <code>java/lang/String</code>. 
     */
    protected abstract boolean requireReference(String internalClassName);
    
    /*
     * @param resourcePath name of resources to read
     */
    protected abstract void requireScript(String resourcePath);
    
    /** Allows subclasses to redefine what field a function representing a
     * class gets assigned. By default it returns the suggested name followed
     * by <code>" = "</code>;
     * 
     * @param className suggested name of the class
     */
    protected String assignClass(String className) {
        return className + " = ";
    }

    /**
     * Converts a given class file to a JavaScript version.
     *
     * @param classFile input stream with code of the .class file
     * @return the initialization code for this class, if any. Otherwise <code>null</code>
     * 
     * @throws IOException if something goes wrong during read or write or translating
     */
    
    public String compile(InputStream classFile) throws IOException {
        this.jc = new ClassData(classFile);
        byte[] arrData = jc.findAnnotationData(true);
        String[] arr = findAnnotation(arrData, jc, 
            "org.apidesign.bck2brwsr.core.ExtraJavaScript", 
            "resource", "processByteCode"
        );
        if (arr != null) {
            requireScript(arr[0]);
            if ("0".equals(arr[1])) {
                return null;
            }
        }
        String[] proto = findAnnotation(arrData, jc, 
            "org.apidesign.bck2brwsr.core.JavaScriptPrototype", 
            "container", "prototype"
        );
        StringArray toInitilize = new StringArray();
        final String className = className(jc);
        out.append("\n\n").append(assignClass(className));
        out.append("function CLS() {");
        out.append("\n  if (!CLS.prototype.$instOf_").append(className).append(") {");
        for (FieldData v : jc.getFields()) {
            if (v.isStatic()) {
                out.append("\n  CLS.").append(v.getName()).append(initField(v));
            }
        }
        if (proto == null) {
            String sc = jc.getSuperClassName(); // with _
            out.append("\n    var p = CLS.prototype = ").
                append(sc.replace('/', '_')).append("(true);");
            out.append("\n    var c = p;");
        } else {
            out.append("\n    var p = CLS.prototype = ").append(proto[1]).append(";");
            out.append("\n    var c = ").append(proto[0]).append(";");
        }
        for (MethodData m : jc.getMethods()) {
            byte[] onlyArr = m.findAnnotationData(true);
            String[] only = findAnnotation(onlyArr, jc, 
                "org.apidesign.bck2brwsr.core.JavaScriptOnly", 
                "name", "value"
            );
            if (only != null) {
                if (only[0] != null && only[1] != null) {
                    out.append("\n    p.").append(only[0]).append(" = ")
                        .append(only[1]).append(";");
                }
                continue;
            }
            if (m.isStatic()) {
                generateStaticMethod("\n    c.", m, toInitilize);
            } else {
                generateInstanceMethod("\n    c.", m);
            }
        }
        out.append("\n    c.constructor = CLS;");
        out.append("\n    c.$instOf_").append(className).append(" = true;");
        for (String superInterface : jc.getSuperInterfaces()) {
            out.append("\n    c.$instOf_").append(superInterface.replace('/', '_')).append(" = true;");
        }
        out.append("\n  }");
        out.append("\n  if (arguments.length === 0) {");
        out.append("\n    if (!(this instanceof CLS)) {");
        out.append("\n      return new CLS();");
        out.append("\n    }");
        for (FieldData v : jc.getFields()) {
            byte[] onlyArr = v.findAnnotationData(true);
            String[] only = findAnnotation(onlyArr, jc, 
                "org.apidesign.bck2brwsr.core.JavaScriptOnly", 
                "name", "value"
            );
            if (only != null) {
                if (only[0] != null && only[1] != null) {
                    out.append("\n    p.").append(only[0]).append(" = ")
                        .append(only[1]).append(";");
                }
                continue;
            }
            if (!v.isStatic()) {
                out.append("\n    this.fld_").
                    append(v.getName()).append(initField(v));
            }
        }
        out.append("\n    return this;");
        out.append("\n  }");
        out.append("\n  return arguments[0] ? new CLS() : CLS.prototype;");
        out.append("\n}");
        StringBuilder sb = new StringBuilder();
        for (String init : toInitilize.toArray()) {
            sb.append("\n").append(init).append("();");
        }
        return sb.toString();
    }
    private void generateStaticMethod(String prefix, MethodData m, StringArray toInitilize) throws IOException {
        if (javaScriptBody(prefix, m, true)) {
            return;
        }
        StringBuilder argsCnt = new StringBuilder();
        final String mn = findMethodName(m, argsCnt);
        out.append(prefix).append(mn).append(" = function");
        if (mn.equals("classV")) {
            toInitilize.add(className(jc) + "(false)." + mn);
        }
        out.append('(');
        String space = "";
        for (int index = 0, i = 0; i < argsCnt.length(); i++) {
            out.append(space);
            out.append("arg").append(String.valueOf(index));
            space = ",";
            final String desc = null;// XXX findDescriptor(args.get(i).getDescriptor());
            if (argsCnt.charAt(i) == '1') {
                index += 2;
            } else {
                index++;
            }
        }
        out.append(") {").append("\n");
        if (m.getCode() != null) {
            int len = m.getMaxLocals();
            for (int i = argsCnt.length(); i < len; i++) {
                out.append("  var ");
                out.append("arg").append(String.valueOf(i)).append(";\n");
            }
            produceCode(m);
        } else {
            out.append("  throw 'no code found for ").append(m.getInternalSig()).append("';\n");
        }
        out.append("};");
    }
    
    private void generateInstanceMethod(String prefix, MethodData m) throws IOException {
        if (javaScriptBody(prefix, m, false)) {
            return;
        }
        StringBuilder argsCnt = new StringBuilder();
        final String mn = findMethodName(m, argsCnt);
        out.append(prefix).append(mn).append(" = function");
        out.append("(arg0");
        String space = ",";
        for (int index = 1, i = 0; i < argsCnt.length(); i++) {
            out.append(space);
            out.append("arg").append(String.valueOf(index));
            if (argsCnt.charAt(i) == '1') {
                index += 2;
            } else {
                index++;
            }
        }
        out.append(") {").append("\n");
        if (m.getCode() != null) {
            int len = m.getMaxLocals();
            for (int i = argsCnt.length(); i < len; i++) {
                out.append("  var ");
                out.append("arg").append(String.valueOf(i + 1)).append(";\n");
            }
            produceCode(m);
        } else {
            out.append("  throw 'no code found for ").append(m.getInternalSig()).append("';\n");
        }
        out.append("};");
    }

    private void produceCode(MethodData m) throws IOException {
        final byte[] byteCodes = m.getCode();
        final StackMapIterator stackMapIterator = m.createStackMapIterator();
        final StackToVariableMapper mapper = new StackToVariableMapper();

        // maxStack includes two stack positions for every pushed long / double
        // so this might generate more stack variables than we need
        final int maxStack = m.getMaxStack();
        if (maxStack > 0) {
            // TODO: generate only used stack variables
            for (int j = 0; j <= Variable.LAST_TYPE; ++j) {
                out.append("\n  var ").append(Variable.getStackVariable(j, 0));
                for (int i = 1; i < maxStack; ++i) {
                    out.append(", ");
                    out.append(Variable.getStackVariable(j, i));
                }
                out.append(';');
            }
        }

        int lastStackFrame = -1;

        out.append("\n  var gt = 0;\n  for(;;) switch(gt) {\n");
        for (int i = 0; i < byteCodes.length; i++) {
            int prev = i;
            stackMapIterator.advanceTo(i);
            if (lastStackFrame != stackMapIterator.getFrameIndex()) {
                lastStackFrame = stackMapIterator.getFrameIndex();
                mapper.syncWithFrameStack(stackMapIterator.getFrameStack());
                out.append("    case " + i).append(": ");
            } else {
                out.append("    /* " + i).append(" */ ");
            }
            final int c = readByte(byteCodes, i);
            switch (c) {
                case opc_aload_0:
                    out.append(mapper.pushA()).append(" = arg0;");
                    break;
                case opc_iload_0:
                    out.append(mapper.pushI()).append(" = arg0;");
                    break;
                case opc_lload_0:
                    out.append(mapper.pushL()).append(" = arg0;");
                    break;
                case opc_fload_0:
                    out.append(mapper.pushF()).append(" = arg0;");
                    break;
                case opc_dload_0:
                    out.append(mapper.pushD()).append(" = arg0;");
                    break;
                case opc_aload_1:
                    out.append(mapper.pushA()).append(" = arg1;");
                    break;
                case opc_iload_1:
                    out.append(mapper.pushI()).append(" = arg1;");
                    break;
                case opc_lload_1:
                    out.append(mapper.pushL()).append(" = arg1;");
                    break;
                case opc_fload_1:
                    out.append(mapper.pushF()).append(" = arg1;");
                    break;
                case opc_dload_1:
                    out.append(mapper.pushD()).append(" = arg1;");
                    break;
                case opc_aload_2:
                    out.append(mapper.pushA()).append(" = arg2;");
                    break;
                case opc_iload_2:
                    out.append(mapper.pushI()).append(" = arg2;");
                    break;
                case opc_lload_2:
                    out.append(mapper.pushL()).append(" = arg2;");
                    break;
                case opc_fload_2:
                    out.append(mapper.pushF()).append(" = arg2;");
                    break;
                case opc_dload_2:
                    out.append(mapper.pushD()).append(" = arg2;");
                    break;
                case opc_aload_3:
                    out.append(mapper.pushA()).append(" = arg3;");
                    break;
                case opc_iload_3:
                    out.append(mapper.pushI()).append(" = arg3;");
                    break;
                case opc_lload_3:
                    out.append(mapper.pushL()).append(" = arg3;");
                    break;
                case opc_fload_3:
                    out.append(mapper.pushF()).append(" = arg3;");
                    break;
                case opc_dload_3:
                    out.append(mapper.pushD()).append(" = arg3;");
                    break;
                case opc_iload: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append(mapper.pushI()).append(" = ")
                       .append("arg").append(indx + ";");
                    break;
                }
                case opc_lload: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append(mapper.pushL()).append(" = ")
                       .append("arg").append(indx + ";");
                    break;
                }
                case opc_fload: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append(mapper.pushF()).append(" = ")
                       .append("arg").append(indx + ";");
                    break;
                }
                case opc_dload: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append(mapper.pushD()).append(" = ")
                       .append("arg").append(indx + ";");
                    break;
                }
                case opc_aload: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append(mapper.pushA()).append(" = ")
                       .append("arg").append(indx + ";");
                    break;
                }
                case opc_istore: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append("arg" + indx).append(" = ")
                       .append(mapper.popI()).append(';');
                    break;
                }
                case opc_lstore: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append("arg" + indx).append(" = ")
                       .append(mapper.popL()).append(';');
                    break;
                }
                case opc_fstore: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append("arg" + indx).append(" = ")
                       .append(mapper.popF()).append(';');
                    break;
                }
                case opc_dstore: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append("arg" + indx).append(" = ")
                       .append(mapper.popD()).append(';');
                    break;
                }
                case opc_astore: {
                    final int indx = readByte(byteCodes, ++i);
                    out.append("arg" + indx).append(" = ")
                       .append(mapper.popA()).append(';');
                    break;
                }
                case opc_astore_0:
                    out.append("arg0 = ").append(mapper.popA()).append(';');
                    break;
                case opc_istore_0:
                    out.append("arg0 = ").append(mapper.popI()).append(';');
                    break;
                case opc_lstore_0:
                    out.append("arg0 = ").append(mapper.popL()).append(';');
                    break;
                case opc_fstore_0:
                    out.append("arg0 = ").append(mapper.popF()).append(';');
                    break;
                case opc_dstore_0:
                    out.append("arg0 = ").append(mapper.popD()).append(';');
                    break;
                case opc_astore_1:
                    out.append("arg1 = ").append(mapper.popA()).append(';');
                    break;
                case opc_istore_1:
                    out.append("arg1 = ").append(mapper.popI()).append(';');
                    break;
                case opc_lstore_1:
                    out.append("arg1 = ").append(mapper.popL()).append(';');
                    break;
                case opc_fstore_1:
                    out.append("arg1 = ").append(mapper.popF()).append(';');
                    break;
                case opc_dstore_1:
                    out.append("arg1 = ").append(mapper.popD()).append(';');
                    break;
                case opc_astore_2:
                    out.append("arg2 = ").append(mapper.popA()).append(';');
                    break;
                case opc_istore_2:
                    out.append("arg2 = ").append(mapper.popI()).append(';');
                    break;
                case opc_lstore_2:
                    out.append("arg2 = ").append(mapper.popL()).append(';');
                    break;
                case opc_fstore_2:
                    out.append("arg2 = ").append(mapper.popF()).append(';');
                    break;
                case opc_dstore_2:
                    out.append("arg2 = ").append(mapper.popD()).append(';');
                    break;
                case opc_astore_3:
                    out.append("arg3 = ").append(mapper.popA()).append(';');
                    break;
                case opc_istore_3:
                    out.append("arg3 = ").append(mapper.popI()).append(';');
                    break;
                case opc_lstore_3:
                    out.append("arg3 = ").append(mapper.popL()).append(';');
                    break;
                case opc_fstore_3:
                    out.append("arg3 = ").append(mapper.popF()).append(';');
                    break;
                case opc_dstore_3:
                    out.append("arg3 = ").append(mapper.popD()).append(';');
                    break;
                case opc_iadd:
                    out.append(mapper.getI(1)).append(" += ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_ladd:
                    out.append(mapper.getL(1)).append(" += ")
                       .append(mapper.popL()).append(';');
                    break;
                case opc_fadd:
                    out.append(mapper.getF(1)).append(" += ")
                       .append(mapper.popF()).append(';');
                    break;
                case opc_dadd:
                    out.append(mapper.getD(1)).append(" += ")
                       .append(mapper.popD()).append(';');
                    break;
                case opc_isub:
                    out.append(mapper.getI(1)).append(" -= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_lsub:
                    out.append(mapper.getL(1)).append(" -= ")
                       .append(mapper.popL()).append(';');
                    break;
                case opc_fsub:
                    out.append(mapper.getF(1)).append(" -= ")
                       .append(mapper.popF()).append(';');
                    break;
                case opc_dsub:
                    out.append(mapper.getD(1)).append(" -= ")
                       .append(mapper.popD()).append(';');
                    break;
                case opc_imul:
                    out.append(mapper.getI(1)).append(" *= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_lmul:
                    out.append(mapper.getL(1)).append(" *= ")
                       .append(mapper.popL()).append(';');
                    break;
                case opc_fmul:
                    out.append(mapper.getF(1)).append(" *= ")
                       .append(mapper.popF()).append(';');
                    break;
                case opc_dmul:
                    out.append(mapper.getD(1)).append(" *= ")
                       .append(mapper.popD()).append(';');
                    break;
                case opc_idiv:
                    out.append(mapper.getI(1)).append(" = ")
                       .append("Math.floor(").append(mapper.getI(1))
                       .append(" / ").append(mapper.popI()).append(");");
                    break;
                case opc_ldiv:
                    out.append(mapper.getL(1)).append(" = ")
                       .append("Math.floor(").append(mapper.getL(1))
                       .append(" / ").append(mapper.popL()).append(");");
                    break;
                case opc_fdiv:
                    out.append(mapper.getF(1)).append(" /= ")
                       .append(mapper.popF()).append(';');
                    break;
                case opc_ddiv:
                    out.append(mapper.getD(1)).append(" /= ")
                       .append(mapper.popD()).append(';');
                    break;
                case opc_irem:
                    out.append(mapper.getI(1)).append(" %= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_lrem:
                    out.append(mapper.getL(1)).append(" %= ")
                       .append(mapper.popL()).append(';');
                    break;
                case opc_frem:
                    out.append(mapper.getF(1)).append(" %= ")
                       .append(mapper.popF()).append(';');
                    break;
                case opc_drem:
                    out.append(mapper.getD(1)).append(" %= ")
                       .append(mapper.popD()).append(';');
                    break;
                case opc_iand:
                    out.append(mapper.getI(1)).append(" &= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_land:
                    out.append(mapper.getL(1)).append(" &= ")
                       .append(mapper.popL()).append(';');
                    break;
                case opc_ior:
                    out.append(mapper.getI(1)).append(" |= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_lor:
                    out.append(mapper.getL(1)).append(" |= ")
                       .append(mapper.popL()).append(';');
                    break;
                case opc_ixor:
                    out.append(mapper.getI(1)).append(" ^= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_lxor:
                    out.append(mapper.getL(1)).append(" ^= ")
                       .append(mapper.popL()).append(';');
                    break;
                case opc_ineg:
                    out.append(mapper.getI(0)).append(" = -")
                       .append(mapper.getI(0)).append(';');
                    break;
                case opc_lneg:
                    out.append(mapper.getL(0)).append(" = -")
                       .append(mapper.getL(0)).append(';');
                    break;
                case opc_fneg:
                    out.append(mapper.getF(0)).append(" = -")
                       .append(mapper.getF(0)).append(';');
                    break;
                case opc_dneg:
                    out.append(mapper.getD(0)).append(" = -")
                       .append(mapper.getD(0)).append(';');
                    break;
                case opc_ishl:
                    out.append(mapper.getI(1)).append(" <<= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_lshl:
                    out.append(mapper.getL(1)).append(" <<= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_ishr:
                    out.append(mapper.getI(1)).append(" >>= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_lshr:
                    out.append(mapper.getL(1)).append(" >>= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_iushr:
                    out.append(mapper.getI(1)).append(" >>>= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_lushr:
                    out.append(mapper.getL(1)).append(" >>>= ")
                       .append(mapper.popI()).append(';');
                    break;
                case opc_iinc: {
                    final int varIndx = readByte(byteCodes, ++i);
                    final int incrBy = byteCodes[++i];
                    if (incrBy == 1) {
                        out.append("arg" + varIndx).append("++;");
                    } else {
                        out.append("arg" + varIndx).append(" += " + incrBy).append(";");
                    }
                    break;
                }
                case opc_return:
                    out.append("return;");
                    break;
                case opc_ireturn:
                    out.append("return ").append(mapper.popI()).append(';');
                    break;
                case opc_lreturn:
                    out.append("return ").append(mapper.popL()).append(';');
                    break;
                case opc_freturn:
                    out.append("return ").append(mapper.popF()).append(';');
                    break;
                case opc_dreturn:
                    out.append("return ").append(mapper.popD()).append(';');
                    break;
                case opc_areturn:
                    out.append("return ").append(mapper.popA()).append(';');
                    break;
                case opc_i2l: {
                    final Variable v = mapper.popI();
                    out.append(mapper.pushL()).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_i2f: {
                    final Variable v = mapper.popI();
                    out.append(mapper.pushF()).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_i2d: {
                    final Variable v = mapper.popI();
                    out.append(mapper.pushD()).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_l2i: {
                    final Variable v = mapper.popL();
                    out.append(mapper.pushI()).append(" = ")
                       .append(v).append(';');
                    break;
                }
                    // max int check?
                case opc_l2f: {
                    final Variable v = mapper.popL();
                    out.append(mapper.pushF()).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_l2d: {
                    final Variable v = mapper.popL();
                    out.append(mapper.pushD()).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_f2d: {
                    final Variable v = mapper.popF();
                    out.append(mapper.pushD()).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_d2f: {
                    final Variable v = mapper.popD();
                    out.append(mapper.pushF()).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_f2i: {
                    final Variable v = mapper.popF();
                    out.append(mapper.pushI()).append(" = ")
                       .append("Math.floor(").append(v).append(");");
                    break;
                }
                case opc_f2l: {
                    final Variable v = mapper.popF();
                    out.append(mapper.pushL()).append(" = ")
                       .append("Math.floor(").append(v).append(");");
                    break;
                }
                case opc_d2i: {
                    final Variable v = mapper.popD();
                    out.append(mapper.pushI()).append(" = ")
                       .append("Math.floor(").append(v).append(");");
                    break;
                }
                case opc_d2l: {
                    final Variable v = mapper.popD();
                    out.append(mapper.pushL()).append(" = ")
                       .append("Math.floor(").append(v).append(");");
                    break;
                }
                case opc_i2b:
                case opc_i2c:
                case opc_i2s:
                    out.append("/* number conversion */");
                    break;
                case opc_aconst_null:
                    out.append(mapper.pushA()).append(" = null;");
                    break;
                case opc_iconst_m1:
                    out.append(mapper.pushI()).append(" = -1;");
                    break;
                case opc_iconst_0:
                    out.append(mapper.pushI()).append(" = 0;");
                    break;
                case opc_dconst_0:
                    out.append(mapper.pushD()).append(" = 0;");
                    break;
                case opc_lconst_0:
                    out.append(mapper.pushL()).append(" = 0;");
                    break;
                case opc_fconst_0:
                    out.append(mapper.pushF()).append(" = 0;");
                    break;
                case opc_iconst_1:
                    out.append(mapper.pushI()).append(" = 1;");
                    break;
                case opc_lconst_1:
                    out.append(mapper.pushL()).append(" = 1;");
                    break;
                case opc_fconst_1:
                    out.append(mapper.pushF()).append(" = 1;");
                    break;
                case opc_dconst_1:
                    out.append(mapper.pushD()).append(" = 1;");
                    break;
                case opc_iconst_2:
                    out.append(mapper.pushI()).append(" = 2;");
                    break;
                case opc_fconst_2:
                    out.append(mapper.pushF()).append(" = 2;");
                    break;
                case opc_iconst_3:
                    out.append(mapper.pushI()).append(" = 3;");
                    break;
                case opc_iconst_4:
                    out.append(mapper.pushI()).append(" = 4;");
                    break;
                case opc_iconst_5:
                    out.append(mapper.pushI()).append(" = 5;");
                    break;
                case opc_ldc: {
                    int indx = readByte(byteCodes, ++i);
                    String v = encodeConstant(indx);
                    int type = constantToVariableType(jc.getTag(indx));
                    out.append(mapper.pushT(type)).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_ldc_w:
                case opc_ldc2_w: {
                    int indx = readIntArg(byteCodes, i);
                    i += 2;
                    String v = encodeConstant(indx);
                    int type = constantToVariableType(jc.getTag(indx));
                    out.append(mapper.pushT(type)).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_lcmp:
                    generateCmp(mapper.popL(), mapper.popL(), mapper.pushI());
                    break;
                case opc_fcmpl:
                case opc_fcmpg:
                    generateCmp(mapper.popF(), mapper.popF(), mapper.pushI());
                    break;
                case opc_dcmpl:
                case opc_dcmpg:
                    generateCmp(mapper.popD(), mapper.popD(), mapper.pushI());
                    break;
                case opc_if_acmpeq:
                    i = generateIf(byteCodes, i, mapper.popA(), mapper.popA(),
                                   "===");
                    break;
                case opc_if_acmpne:
                    i = generateIf(byteCodes, i, mapper.popA(), mapper.popA(),
                                   "!=");
                    break;
                case opc_if_icmpeq: {
                    i = generateIf(byteCodes, i, mapper.popI(), mapper.popI(),
                                   "==");
                    break;
                }
                case opc_ifeq: {
                    int indx = i + readIntArg(byteCodes, i);
                    out.append("if (").append(mapper.popI())
                                      .append(" == 0) { gt = " + indx);
                    out.append("; continue; }");
                    i += 2;
                    break;
                }
                case opc_ifne: {
                    int indx = i + readIntArg(byteCodes, i);
                    out.append("if (").append(mapper.popI())
                                      .append(" != 0) { gt = " + indx);
                    out.append("; continue; }");
                    i += 2;
                    break;
                }
                case opc_iflt: {
                    int indx = i + readIntArg(byteCodes, i);
                    out.append("if (").append(mapper.popI())
                                      .append(" < 0) { gt = " + indx);
                    out.append("; continue; }");
                    i += 2;
                    break;
                }
                case opc_ifle: {
                    int indx = i + readIntArg(byteCodes, i);
                    out.append("if (").append(mapper.popI())
                                      .append(" <= 0) { gt = " + indx);
                    out.append("; continue; }");
                    i += 2;
                    break;
                }
                case opc_ifgt: {
                    int indx = i + readIntArg(byteCodes, i);
                    out.append("if (").append(mapper.popI())
                                      .append(" > 0) { gt = " + indx);
                    out.append("; continue; }");
                    i += 2;
                    break;
                }
                case opc_ifge: {
                    int indx = i + readIntArg(byteCodes, i);
                    out.append("if (").append(mapper.popI())
                                      .append(" >= 0) { gt = " + indx);
                    out.append("; continue; }");
                    i += 2;
                    break;
                }
                case opc_ifnonnull: {
                    int indx = i + readIntArg(byteCodes, i);
                    out.append("if (").append(mapper.popA())
                                      .append(" !== null) { gt = " + indx);
                    out.append("; continue; }");
                    i += 2;
                    break;
                }
                case opc_ifnull: {
                    int indx = i + readIntArg(byteCodes, i);
                    out.append("if (").append(mapper.popA())
                                      .append(" === null) { gt = " + indx);
                    out.append("; continue; }");
                    i += 2;
                    break;
                }
                case opc_if_icmpne:
                    i = generateIf(byteCodes, i, mapper.popI(), mapper.popI(),
                                   "!=");
                    break;
                case opc_if_icmplt:
                    i = generateIf(byteCodes, i, mapper.popI(), mapper.popI(),
                                   "<");
                    break;
                case opc_if_icmple:
                    i = generateIf(byteCodes, i, mapper.popI(), mapper.popI(),
                                   "<=");
                    break;
                case opc_if_icmpgt:
                    i = generateIf(byteCodes, i, mapper.popI(), mapper.popI(),
                                   ">");
                    break;
                case opc_if_icmpge:
                    i = generateIf(byteCodes, i, mapper.popI(), mapper.popI(),
                                   ">=");
                    break;
                case opc_goto: {
                    int indx = i + readIntArg(byteCodes, i);
                    out.append("gt = " + indx).append("; continue;");
                    i += 2;
                    break;
                }
                case opc_lookupswitch: {
                    int table = i / 4 * 4 + 4;
                    int dflt = i + readInt4(byteCodes, table);
                    table += 4;
                    int n = readInt4(byteCodes, table);
                    table += 4;
                    out.append("switch (").append(mapper.popI()).append(") {\n");
                    while (n-- > 0) {
                        int cnstnt = readInt4(byteCodes, table);
                        table += 4;
                        int offset = i + readInt4(byteCodes, table);
                        table += 4;
                        out.append("  case " + cnstnt).append(": gt = " + offset).append("; continue;\n");
                    }
                    out.append("  default: gt = " + dflt).append("; continue;\n}");
                    i = table - 1;
                    break;
                }
                case opc_tableswitch: {
                    int table = i / 4 * 4 + 4;
                    int dflt = i + readInt4(byteCodes, table);
                    table += 4;
                    int low = readInt4(byteCodes, table);
                    table += 4;
                    int high = readInt4(byteCodes, table);
                    table += 4;
                    out.append("switch (").append(mapper.popI()).append(") {\n");
                    while (low <= high) {
                        int offset = i + readInt4(byteCodes, table);
                        table += 4;
                        out.append("  case " + low).append(": gt = " + offset).append("; continue;\n");
                        low++;
                    }
                    out.append("  default: gt = " + dflt).append("; continue;\n}");
                    i = table - 1;
                    break;
                }
                case opc_invokeinterface: {
                    i = invokeVirtualMethod(byteCodes, i, mapper) + 2;
                    break;
                }
                case opc_invokevirtual:
                    i = invokeVirtualMethod(byteCodes, i, mapper);
                    break;
                case opc_invokespecial:
                    i = invokeStaticMethod(byteCodes, i, mapper, false);
                    break;
                case opc_invokestatic:
                    i = invokeStaticMethod(byteCodes, i, mapper, true);
                    break;
                case opc_new: {
                    int indx = readIntArg(byteCodes, i);
                    String ci = jc.getClassName(indx);
                    out.append(mapper.pushA()).append(" = ");
                    out.append("new ").append(ci.replace('/','_'));
                    out.append(';');
                    addReference(ci);
                    i += 2;
                    break;
                }
                case opc_newarray: {
                    ++i; // skip type of array
                    final Variable v = mapper.popI();
                    out.append(mapper.pushA()).append(" = ")
                       .append("new Array(").append(v).append(").fillNulls();");
                    break;
                }
                case opc_anewarray: {
                    i += 2; // skip type of array
                    final Variable v = mapper.popI();
                    out.append(mapper.pushA()).append(" = ")
                       .append("new Array(").append(v).append(").fillNulls();");
                    break;
                }
                case opc_multianewarray: {
                    i += 2;
                    int dim = readByte(byteCodes, ++i);
                    out.append("{ var a0 = new Array(").append(mapper.popI())
                       .append(").fillNulls();");
                    for (int d = 1; d < dim; d++) {
                        out.append("\n  var l" + d).append(" = ")
                           .append(mapper.popI()).append(';');
                        out.append("\n  for (var i" + d).append (" = 0; i" + d).
                            append(" < a" + (d - 1)).
                            append(".length; i" + d).append("++) {");
                        out.append("\n    var a" + d).
                            append (" = new Array(l" + d).append(").fillNulls();");
                        out.append("\n    a" + (d - 1)).append("[i" + d).append("] = a" + d).
                            append(";");
                    }
                    for (int d = 1; d < dim; d++) {
                        out.append("\n  }");
                    }
                    out.append("\n").append(mapper.pushA()).append(" = a0; }");
                    break;
                }
                case opc_arraylength: {
                    final Variable v = mapper.popA();
                    out.append(mapper.pushI()).append(" = ")
                       .append(v).append(".length;");
                    break;
                }
                case opc_lastore: {
                    out.append(mapper.getA(2)).append('[')
                                              .append(mapper.getI(1))
                                              .append(']')
                       .append(" = ").append(mapper.getL(0)).append(';');
                    mapper.pop(3);
                    break;
                }
                case opc_fastore: {
                    out.append(mapper.getA(2)).append('[')
                                              .append(mapper.getI(1))
                                              .append(']')
                       .append(" = ").append(mapper.getF(0)).append(';');
                    mapper.pop(3);
                    break;
                }
                case opc_dastore: {
                    out.append(mapper.getA(2)).append('[')
                                              .append(mapper.getI(1))
                                              .append(']')
                       .append(" = ").append(mapper.getD(0)).append(';');
                    mapper.pop(3);
                    break;
                }
                case opc_aastore: {
                    out.append(mapper.getA(2)).append('[')
                                              .append(mapper.getI(1))
                                              .append(']')
                       .append(" = ").append(mapper.getA(0)).append(';');
                    mapper.pop(3);
                    break;
                }
                case opc_iastore:
                case opc_bastore:
                case opc_castore:
                case opc_sastore: {
                    out.append(mapper.getA(2)).append('[')
                                              .append(mapper.getI(1))
                                              .append(']')
                       .append(" = ").append(mapper.getI(0)).append(';');
                    mapper.pop(3);
                    break;
                }
                case opc_laload: {
                    final Variable vidx = mapper.popI();
                    final Variable varr = mapper.popA();
                    out.append(mapper.pushL()).append(" = ")
                       .append(varr).append('[').append(vidx).append("];");
                    break;
                }
                case opc_faload: {
                    final Variable vidx = mapper.popI();
                    final Variable varr = mapper.popA();
                    out.append(mapper.pushF()).append(" = ")
                       .append(varr).append('[').append(vidx).append("];");
                    break;
                }
                case opc_daload: {
                    final Variable vidx = mapper.popI();
                    final Variable varr = mapper.popA();
                    out.append(mapper.pushD()).append(" = ")
                       .append(varr).append('[').append(vidx).append("];");
                    break;
                }
                case opc_aaload: {
                    final Variable vidx = mapper.popI();
                    final Variable varr = mapper.popA();
                    out.append(mapper.pushA()).append(" = ")
                       .append(varr).append('[').append(vidx).append("];");
                    break;
                }
                case opc_iaload:
                case opc_baload:
                case opc_caload:
                case opc_saload: {
                    final Variable vidx = mapper.popI();
                    final Variable varr = mapper.popA();
                    out.append(mapper.pushI()).append(" = ")
                       .append(varr).append('[').append(vidx).append("];");
                    break;
                }
                case opc_pop:
                case opc_pop2:
                    mapper.pop(1);
                    out.append("/* pop */");
                    break;
                case opc_dup: {
                    final Variable v = mapper.get(0);
                    out.append(mapper.pushT(v.getType())).append(" = ")
                       .append(v).append(';');
                    break;
                }
                case opc_dup2: {
                    if (mapper.get(0).isCategory2()) {
                        final Variable v = mapper.get(0);
                        out.append(mapper.pushT(v.getType())).append(" = ")
                           .append(v).append(';');
                    } else {
                        final Variable v1 = mapper.get(0);
                        final Variable v2 = mapper.get(1);
                        out.append("{ ");
                        out.append(mapper.pushT(v2.getType())).append(" = ")
                           .append(v2).append("; ");
                        out.append(mapper.pushT(v1.getType())).append(" = ")
                           .append(v1).append("; ");
                        out.append('}');
                    }
                    break;
                }
                case opc_dup_x1: {
                    final Variable vi1 = mapper.pop();
                    final Variable vi2 = mapper.pop();
                    final Variable vo3 = mapper.pushT(vi1.getType());
                    final Variable vo2 = mapper.pushT(vi2.getType());
                    final Variable vo1 = mapper.pushT(vi1.getType());

                    out.append("{ ");
                    out.append(vo1).append(" = ").append(vi1).append("; ");
                    out.append(vo2).append(" = ").append(vi2).append("; ");
                    out.append(vo3).append(" = ").append(vo1).append("; ");
                    out.append('}');
                    break;
                }
                case opc_dup_x2: {
                    if (mapper.get(1).isCategory2()) {
                        final Variable vi1 = mapper.pop();
                        final Variable vi2 = mapper.pop();
                        final Variable vo3 = mapper.pushT(vi1.getType());
                        final Variable vo2 = mapper.pushT(vi2.getType());
                        final Variable vo1 = mapper.pushT(vi1.getType());

                        out.append("{ ");
                        out.append(vo1).append(" = ").append(vi1).append("; ");
                        out.append(vo2).append(" = ").append(vi2).append("; ");
                        out.append(vo3).append(" = ").append(vo1).append("; ");
                        out.append('}');
                    } else {
                        final Variable vi1 = mapper.pop();
                        final Variable vi2 = mapper.pop();
                        final Variable vi3 = mapper.pop();
                        final Variable vo4 = mapper.pushT(vi1.getType());
                        final Variable vo3 = mapper.pushT(vi3.getType());
                        final Variable vo2 = mapper.pushT(vi2.getType());
                        final Variable vo1 = mapper.pushT(vi1.getType());

                        out.append("{ ");
                        out.append(vo1).append(" = ").append(vi1).append("; ");
                        out.append(vo2).append(" = ").append(vi2).append("; ");
                        out.append(vo3).append(" = ").append(vi3).append("; ");
                        out.append(vo4).append(" = ").append(vo1).append("; ");
                        out.append('}');
                    }
                    break;
                }
                case opc_bipush:
                    out.append(mapper.pushI()).append(" = ")
                       .append(Integer.toString(byteCodes[++i])).append(';');
                    break;
                case opc_sipush:
                    out.append(mapper.pushI()).append(" = ")
                       .append(Integer.toString(readIntArg(byteCodes, i)))
                       .append(';');
                    i += 2;
                    break;
                case opc_getfield: {
                    int indx = readIntArg(byteCodes, i);
                    String[] fi = jc.getFieldInfoName(indx);
                    final int type = fieldToVariableType(fi[2].charAt(0));
                    final Variable v = mapper.popA();
                    out.append(mapper.pushT(type)).append(" = ")
                       .append(v).append(".fld_").append(fi[1]).append(';');
                    i += 2;
                    break;
                }
                case opc_getstatic: {
                    int indx = readIntArg(byteCodes, i);
                    String[] fi = jc.getFieldInfoName(indx);
                    final int type = fieldToVariableType(fi[2].charAt(0));
                    out.append(mapper.pushT(type)).append(" = ")
                       .append(fi[0].replace('/', '_'))
                       .append('.').append(fi[1]).append(';');
                    i += 2;
                    addReference(fi[0]);
                    break;
                }
                case opc_putfield: {
                    int indx = readIntArg(byteCodes, i);
                    String[] fi = jc.getFieldInfoName(indx);
                    final int type = fieldToVariableType(fi[2].charAt(0));
                    final Variable v = mapper.popT(type);
                    out.append(mapper.popA()).append(".fld_").append(fi[1])
                       .append(" = ").append(v).append(';');
                    i += 2;
                    break;
                }
                case opc_putstatic: {
                    int indx = readIntArg(byteCodes, i);
                    String[] fi = jc.getFieldInfoName(indx);
                    final int type = fieldToVariableType(fi[2].charAt(0));
                    out.append(fi[0].replace('/', '_'));
                    out.append('.').append(fi[1]).append(" = ")
                       .append(mapper.popT(type)).append(';');
                    i += 2;
                    addReference(fi[0]);
                    break;
                }
                case opc_checkcast: {
                    int indx = readIntArg(byteCodes, i);
                    final String type = jc.getClassName(indx);
                    if (!type.startsWith("[")) {
                        // no way to check arrays right now
                        out.append("if (").append(mapper.getA(0))
                           .append(".$instOf_").append(type.replace('/', '_'))
                           .append(" != 1) throw {};"); // XXX proper exception
                    }
                    i += 2;
                    break;
                }
                case opc_instanceof: {
                    int indx = readIntArg(byteCodes, i);
                    final String type = jc.getClassName(indx);
                    final Variable v = mapper.popA();
                    out.append(mapper.pushI()).append(" = ")
                       .append(v).append(".$instOf_")
                                 .append(type.replace('/', '_'))
                       .append(" ? 1 : 0;");
                    i += 2;
                    break;
                }
                case opc_athrow: {
                    final Variable v = mapper.popA();
                    mapper.clear();

                    out.append("{ ");
                    out.append(mapper.pushA()).append(" = ")
                       .append(v).append("; ");
                    out.append("throw ").append(v).append("; ");
                    out.append('}');

                    break;
                }

                case opc_monitorenter: {
                    out.append("/* monitor enter */");
                    mapper.popA();
                    break;
                }

                case opc_monitorexit: {
                    out.append("/* monitor exit */");
                    mapper.popA();
                    break;
                }

                default: {
                    out.append("throw 'unknown bytecode " + c + "';");
                }
                    
            }
            out.append(" //");
            for (int j = prev; j <= i; j++) {
                out.append(" ");
                final int cc = readByte(byteCodes, j);
                out.append(Integer.toString(cc));
            }
            out.append("\n");
        }
        out.append("  }\n");
    }

    private int generateIf(byte[] byteCodes, int i,
                           final Variable v2, final Variable v1,
                           final String test) throws IOException {
        int indx = i + readIntArg(byteCodes, i);
        out.append("if (").append(v1)
           .append(' ').append(test).append(' ')
           .append(v2).append(") { gt = " + indx)
           .append("; continue; }");
        return i + 2;
    }

    private void generateCmp(Variable v2, Variable v1, Variable vr)
            throws IOException {
        out.append(vr).append(" = ")
           .append('(').append(v1).append(" == ").append(v2).append(')')
           .append(" ? 0")
           .append(" : ((").append(v1).append(" < ").append(v2).append(')')
           .append(" ? -1 : 1);");
    }

    private int readIntArg(byte[] byteCodes, int offsetInstruction) {
        final int indxHi = byteCodes[offsetInstruction + 1] << 8;
        final int indxLo = byteCodes[offsetInstruction + 2];
        return (indxHi & 0xffffff00) | (indxLo & 0xff);
    }
    private int readInt4(byte[] byteCodes, int offsetInstruction) {
        final int d = byteCodes[offsetInstruction + 0] << 24;
        final int c = byteCodes[offsetInstruction + 1] << 16;
        final int b = byteCodes[offsetInstruction + 2] << 8;
        final int a = byteCodes[offsetInstruction + 3];
        return (d & 0xff000000) | (c & 0xff0000) | (b & 0xff00) | (a & 0xff);
    }
    private int readByte(byte[] byteCodes, int offsetInstruction) {
        return byteCodes[offsetInstruction] & 0xff;
    }
    
    private static void countArgs(String descriptor, char[] returnType, StringBuilder sig, StringBuilder cnt) {
        int i = 0;
        Boolean count = null;
        boolean array = false;
        int firstPos = sig.length();
        while (i < descriptor.length()) {
            char ch = descriptor.charAt(i++);
            switch (ch) {
                case '(':
                    count = true;
                    continue;
                case ')':
                    count = false;
                    continue;
                case 'A':
                    array = true;
                    break;
                case 'B': 
                case 'C': 
                case 'D': 
                case 'F': 
                case 'I': 
                case 'J': 
                case 'S': 
                case 'Z': 
                    if (count) {
                        if (array) {
                            sig.append('A');
                        }
                        sig.append(ch);
                        if (ch == 'J' || ch == 'D') {
                            cnt.append('1');
                        } else {
                            cnt.append('0');
                        }
                    } else {
                        sig.insert(firstPos, ch);
                        if (array) {
                            returnType[0] = '[';
                            sig.insert(firstPos, 'A');
                        } else {
                            returnType[0] = ch;
                        }
                    }
                    array = false;
                    continue;
                case 'V': 
                    assert !count;
                    returnType[0] = 'V';
                    sig.insert(firstPos, 'V');
                    continue;
                case 'L':
                    int next = descriptor.indexOf(';', i);
                    if (count) {
                        if (array) {
                            sig.append('A');
                        }
                        sig.append(ch);
                        sig.append(descriptor.substring(i, next).replace('/', '_'));
                        cnt.append('0');
                    } else {
                        sig.insert(firstPos, descriptor.substring(i, next).replace('/', '_'));
                        sig.insert(firstPos, ch);
                        if (array) {
                            sig.insert(firstPos, 'A');
                        }
                        returnType[0] = 'L';
                    }
                    i = next + 1;
                    continue;
                case '[':
                    //arrays++;
                    continue;
                default:
                    break; // invalid character
            }
        }
    }

    private String findMethodName(MethodData m, StringBuilder cnt) {
        StringBuilder name = new StringBuilder();
        if ("<init>".equals(m.getName())) { // NOI18N
            name.append("cons"); // NOI18N
        } else if ("<clinit>".equals(m.getName())) { // NOI18N
            name.append("class"); // NOI18N
        } else {
            name.append(m.getName());
        } 
        
        countArgs(findDescriptor(m.getInternalSig()), new char[1], name, cnt);
        return name.toString();
    }

    private String findMethodName(String[] mi, StringBuilder cnt, char[] returnType) {
        StringBuilder name = new StringBuilder();
        String descr = mi[2];//mi.getDescriptor();
        String nm= mi[1];
        if ("<init>".equals(nm)) { // NOI18N
            name.append("cons"); // NOI18N
        } else {
            name.append(nm);
        }
        countArgs(findDescriptor(descr), returnType, name, cnt);
        return name.toString();
    }

    private int invokeStaticMethod(byte[] byteCodes, int i, final StackToVariableMapper mapper, boolean isStatic)
    throws IOException {
        int methodIndex = readIntArg(byteCodes, i);
        String[] mi = jc.getFieldInfoName(methodIndex);
        char[] returnType = { 'V' };
        StringBuilder cnt = new StringBuilder();
        String mn = findMethodName(mi, cnt, returnType);

        final int numArguments = isStatic ? cnt.length() : cnt.length() + 1;
        final Variable[] vars = new Variable[numArguments];

        for (int j = numArguments - 1; j >= 0; --j) {
            vars[j] = mapper.pop();
        }

        if (returnType[0] != 'V') {
            out.append(mapper.pushT(fieldToVariableType(returnType[0])))
               .append(" = ");
        }

        final String in = mi[0];
        out.append(in.replace('/', '_'));
        out.append("(false).");
        out.append(mn);
        out.append('(');
        if (numArguments > 0) {
            out.append(vars[0]);
            for (int j = 1; j < numArguments; ++j) {
                out.append(", ");
                out.append(vars[j]);
            }
        }
        out.append(");");
        i += 2;
        addReference(in);
        return i;
    }
    private int invokeVirtualMethod(byte[] byteCodes, int i, final StackToVariableMapper mapper)
    throws IOException {
        int methodIndex = readIntArg(byteCodes, i);
        String[] mi = jc.getFieldInfoName(methodIndex);
        char[] returnType = { 'V' };
        StringBuilder cnt = new StringBuilder();
        String mn = findMethodName(mi, cnt, returnType);

        final int numArguments = cnt.length() + 1;
        final Variable[] vars = new Variable[numArguments];

        for (int j = numArguments - 1; j >= 0; --j) {
            vars[j] = mapper.pop();
        }

        if (returnType[0] != 'V') {
            out.append(mapper.pushT(fieldToVariableType(returnType[0])))
               .append(" = ");
        }

        out.append(vars[0]).append('.');
        out.append(mn);
        out.append('(');
        out.append(vars[0]);
        for (int j = 1; j < numArguments; ++j) {
            out.append(", ");
            out.append(vars[j]);
        }
        out.append(");");
        i += 2;
        return i;
    }

    private void addReference(String cn) throws IOException {
        if (requireReference(cn)) {
            out.append(" /* needs ").append(cn).append(" */");
        }
    }

    private void outType(String d, StringBuilder out) {
        int arr = 0;
        while (d.charAt(0) == '[') {
            out.append('A');
            d = d.substring(1);
        }
        if (d.charAt(0) == 'L') {
            assert d.charAt(d.length() - 1) == ';';
            out.append(d.replace('/', '_').substring(0, d.length() - 1));
        } else {
            out.append(d);
        }
    }

    private String encodeConstant(int entryIndex) {
        String s = jc.stringValue(entryIndex, true);
        return s;
    }

    private String findDescriptor(String d) {
        return d.replace('[', 'A');
    }

    private boolean javaScriptBody(String prefix, MethodData m, boolean isStatic) throws IOException {
        byte[] arr = m.findAnnotationData(true);
        if (arr == null) {
            return false;
        }
        final String jvmType = "Lorg/apidesign/bck2brwsr/core/JavaScriptBody;";
        class P extends AnnotationParser {
            int cnt;
            String[] args = new String[30];
            String body;
            
            @Override
            protected void visitAttr(String type, String attr, String value) {
                if (type.equals(jvmType)) {
                    if ("body".equals(attr)) {
                        body = value;
                    } else if ("args".equals(attr)) {
                        args[cnt++] = value;
                    } else {
                        throw new IllegalArgumentException(attr);
                    }
                }
            }
        }
        P p = new P();
        p.parse(arr, jc);
        if (p.body == null) {
            return false;
        }
        StringBuilder cnt = new StringBuilder();
        out.append(prefix).append(findMethodName(m, cnt));
        out.append(" = function(");
        String space;
        int index;
        if (!isStatic) {                
            out.append(p.args[0]);
            space = ",";
            index = 1;
        } else {
            space = "";
            index = 0;
        }
        for (int i = 0; i < cnt.length(); i++) {
            out.append(space);
            out.append(p.args[index]);
            index++;
            space = ",";
        }
        out.append(") {").append("\n");
        out.append(p.body);
        out.append("\n}\n");
        return true;
    }
    private static String className(ClassData jc) {
        //return jc.getName().getInternalName().replace('/', '_');
        return jc.getClassName().replace('/', '_');
    }
    
    private static String[] findAnnotation(
        byte[] arr, ClassData cd, final String className, 
        final String... attrNames
    ) throws IOException {
        if (arr == null) {
            return null;
        }
        final String[] values = new String[attrNames.length];
        final boolean[] found = { false };
        final String jvmType = "L" + className.replace('.', '/') + ";";
        AnnotationParser ap = new AnnotationParser() {
            @Override
            protected void visitAttr(String type, String attr, String value) {
                if (type.equals(jvmType)) {
                    found[0] = true;
                    for (int i = 0; i < attrNames.length; i++) {
                        if (attrNames[i].equals(attr)) {
                            values[i] = value;
                        }
                    }
                }
            }
            
        };
        ap.parse(arr, cd);
        return found[0] ? values : null;
    }

    private CharSequence initField(FieldData v) {
        final String is = v.getInternalSig();
        if (is.length() == 1) {
            switch (is.charAt(0)) {
                case 'S':
                case 'J':
                case 'B':
                case 'Z':
                case 'C':
                case 'I': return " = 0;";
                case 'F': 
                case 'D': return " = 0.0;";
                default:
                    throw new IllegalStateException(is);
            }
        }
        return " = null;";
    }

    private static int constantToVariableType(final byte constantTag) {
        switch (constantTag) {
            case CONSTANT_INTEGER:
                return Variable.TYPE_INT;
            case CONSTANT_FLOAT:
                return Variable.TYPE_FLOAT;
            case CONSTANT_LONG:
                return Variable.TYPE_LONG;
            case CONSTANT_DOUBLE:
                return Variable.TYPE_DOUBLE;

            case CONSTANT_CLASS:
            case CONSTANT_UTF8:
            case CONSTANT_UNICODE:
            case CONSTANT_STRING:
                return Variable.TYPE_REF;

            case CONSTANT_FIELD:
            case CONSTANT_METHOD:
            case CONSTANT_INTERFACEMETHOD:
            case CONSTANT_NAMEANDTYPE:
                /* unclear how to handle for now */
            default:
                throw new IllegalStateException("Unhandled constant tag");
        }
    }

    private static int fieldToVariableType(final char fieldType) {
        switch (fieldType) {
            case 'B':
            case 'C':
            case 'S':
            case 'Z':
            case 'I':
                return Variable.TYPE_INT;
            case 'J':
                return Variable.TYPE_LONG;
            case 'F':
                return Variable.TYPE_FLOAT;
            case 'D':
                return Variable.TYPE_DOUBLE;
            case 'L':
            case '[':
                return Variable.TYPE_REF;

            default:
                throw new IllegalStateException("Unhandled field type");
        }
    }
}
